#ifndef PICOLED_BUFFER_H
#define PICOLED_BUFFER_H

#include <cstddef>
#include <cstdint>

namespace picoled
{

class fixed
{
public:
  fixed();

  fixed(float value);
  operator int16_t() const;

  fixed operator-() const;
  fixed operator~() const;

  fixed operator+(const fixed& other) const;
  fixed operator-(const fixed& other) const;
  fixed operator*(const fixed& other) const;
  fixed operator/(const fixed& other) const;

  fixed& operator+=(const fixed& other);
  fixed& operator-=(const fixed& other);
  fixed& operator*=(const fixed& other);
  fixed& operator/=(const fixed& other);

  bool operator<(const fixed& other) const;
  bool operator<=(const fixed& other) const;
  bool operator>(const fixed& other) const;
  bool operator>=(const fixed& other) const;
  bool operator==(const fixed& other) const;

  static fixed min(fixed a, fixed b);
  static fixed max(fixed a, fixed b);

private:
  int32_t value;
};

// behavior when out of bounds
enum class buffer_read_mode
{
  empty, // use default constructor for T
  wrap,    // use modular arithmetic to put back in bounds
  clamp    // pick from the corresponding edge
};

template <typename T>
class buffer
{
public:
  buffer(size_t width, size_t height) :
    width(width),
    height(height),
    data(new T[width * height])
  {
  }
  ~buffer()
  {
    delete[] data;
  }

  buffer(const buffer&) = delete;
  buffer(buffer&&) = delete;

  buffer& operator=(const buffer&) = delete;
  buffer& operator=(buffer&&) = delete;

  size_t get_width() const { return width; }
  size_t get_height() const { return height; }

  void write(int x, int y, const T& pixel)
  {
    if (x < 0 || x >= width || y < 0 || y >= height) { return; }

    data[x + y * width] = pixel;
  }

  T read_uv(fixed x, fixed y, buffer_read_mode mode = buffer_read_mode::clamp) const
  {
    return read(x * fixed(width), y * fixed(height), mode);
  }

  T read(int x, int y, buffer_read_mode mode = buffer_read_mode::clamp) const
  {
    switch (mode)
    {
    case buffer_read_mode::empty:
      return read_default(x, y);

    case buffer_read_mode::wrap:
      return read_wrap(x, y);

    case buffer_read_mode::clamp:
      return read_clamp(x, y);
    }

    return T();
  }

  const uint8_t* data_pointer() const
  {
    return reinterpret_cast<const uint8_t*>(data);
  }

  void clear()
  {
    for (int i = 0; i < width * height; i++)
    {
      data[i] = T();
    }
  }

private:
  size_t width;
  size_t height;
  T* data;

  T read_default(int x, int y) const
  {
    if (x < 0 || x >= width || y < 0 || y >= height) { return T(); }
    return data[x + y * width];
  }

  T read_wrap(int x, int y) const
  {
    if (x < 0)
    {
      x = width - ((-x) % width);
    }
    else if (x >= width)
    {
      x = x % width;
    }

    if (y < 0)
    {
      y = height - ((-y) % height);
    }
    else if (y >= height)
    {
      y = y % height;
    }

    return data[x + y * width];
  }

  T read_clamp(int x, int y) const
  {
    if (x < 0) { x = 0; }
    else if (x >= width) { x = width - 1; }

    if (y < 0) { y = 0; }
    else if (y >= height) { y = height - 1; }

    return data[x + y * width];
  }
};

}

#endif//PICOLED_BUFFER_H
