#include "picoled/gfx/2D.h"
#include "picoled/buffer.h"

#include <cmath>
#include <algorithm>

#include <pico/stdio.h>

namespace picoled::gfx
{

triangle triangle::sort_y() const
{
  // sort from top to bottom (lowest -> highest y)
  triangle t = *this;

  // sort a
  if (t.a.position.y > t.b.position.y) { std::swap(t.a, t.b); }
  if (t.b.position.y > t.c.position.y) { std::swap(t.b, t.c); }

  // sort b and c
  if (t.a.position.y > t.b.position.y) { std::swap(t.a, t.b); }

  return t;
}

bool triangle::left_sided() const
{
  // *this must be sorted by y

  const vertex2* leftmost = &a;
  if (b.position.x < leftmost->position.x) { leftmost = &b; }
  if (c.position.x < leftmost->position.x) { leftmost = &c; }

  return leftmost == &a || leftmost == &c;
}

vec3 triangle::barycentric(const vec2& p) const
{
  const vec2 ba = b.position - a.position;
  const vec2 ca = c.position - a.position;
  const vec2 pa = p - a.position;

  const fixed denominator_inv = fixed(1) / (ba.x * ca.y - ca.x * ba.y);

  const fixed a = (pa.x * ca.y - ca.x * pa.y) * denominator_inv;
  const fixed b = (ba.x * pa.y - pa.x * ba.y) * denominator_inv;

  return { a, b, fixed(1) - a - b };
}

void gfx2d::draw_triangle(const triangle& tri, const buffer<pixel>& texture, bool ignore_winding)
{
  const triangle sorted_tri = tri.sort_y();
  const bool left_sided = sorted_tri.left_sided();

  // tall segment
  const fixed inv_slope_ac = (sorted_tri.c.position.x - sorted_tri.a.position.x) / (sorted_tri.c.position.y - sorted_tri.a.position.y);

  // top other side segment
  const fixed inv_slope_ab = (sorted_tri.b.position.x - sorted_tri.a.position.x) / (sorted_tri.b.position.y - sorted_tri.a.position.y);

  const fixed upper_left_slope  = left_sided ? inv_slope_ac : inv_slope_ab;
  const fixed upper_right_slope = left_sided ? inv_slope_ab : inv_slope_ac;

  // render upper triangle
  for (fixed y = 0; y < sorted_tri.b.position.y - sorted_tri.a.position.y + fixed(0.2f); y += 1)
  {
    const fixed left  = y * upper_left_slope  + sorted_tri.a.position.x;
    const fixed right = y * upper_right_slope + sorted_tri.a.position.x;

    const vec3 barycentric_left  = tri.barycentric({ left,  y + sorted_tri.a.position.y });
    const vec3 barycentric_right = tri.barycentric({ right, y + sorted_tri.a.position.y });

    const fixed u_left = barycentric_left % vec3{ tri.a.texture_coordinates.x, tri.b.texture_coordinates.x, tri.c.texture_coordinates.x };
    const fixed v_left = barycentric_left % vec3{ tri.a.texture_coordinates.y, tri.b.texture_coordinates.y, tri.c.texture_coordinates.y };

    const fixed u_right = barycentric_right % vec3{ tri.a.texture_coordinates.x, tri.b.texture_coordinates.x, tri.c.texture_coordinates.x };
    const fixed v_right = barycentric_right % vec3{ tri.a.texture_coordinates.y, tri.b.texture_coordinates.y, tri.c.texture_coordinates.y };

    const fixed distance_inv = fixed(1) / (right - left);
    const fixed u_delta = (u_right - u_left) * distance_inv;
    const fixed v_delta = (v_right - v_left) * distance_inv;

    fixed u = u_left;
    fixed v = v_left;

    for (fixed x = left; x < right; x += 1)
    {
      screen.write(x, y + sorted_tri.a.position.y, texture.read_uv(u, v));
      u += u_delta;
      v += v_delta;
    }
  }

  // bottom side other segment
  const fixed inv_slope_bc = (sorted_tri.c.position.x - sorted_tri.b.position.x) / (sorted_tri.c.position.y - sorted_tri.b.position.y);

  const fixed lower_left_slope  = left_sided ? inv_slope_ac : inv_slope_bc;
  const fixed lower_right_slope = left_sided ? inv_slope_bc : inv_slope_ac;

  // render lower triangle
  for (fixed y = 0; y < sorted_tri.c.position.y - sorted_tri.b.position.y + fixed(0.2f); y += 1)
  {
    // going from bottom to top, multiply y by -1
    const fixed left  = -y * lower_left_slope  + sorted_tri.c.position.x;
    const fixed right = -y * lower_right_slope + sorted_tri.c.position.x;

    const vec3 barycentric_left  = tri.barycentric({ left,  -y + sorted_tri.c.position.y });
    const vec3 barycentric_right = tri.barycentric({ right, -y + sorted_tri.c.position.y });

    const fixed u_left = barycentric_left % vec3{ tri.a.texture_coordinates.x, tri.b.texture_coordinates.x, tri.c.texture_coordinates.x };
    const fixed v_left = barycentric_left % vec3{ tri.a.texture_coordinates.y, tri.b.texture_coordinates.y, tri.c.texture_coordinates.y };

    const fixed u_right = barycentric_right % vec3{ tri.a.texture_coordinates.x, tri.b.texture_coordinates.x, tri.c.texture_coordinates.x };
    const fixed v_right = barycentric_right % vec3{ tri.a.texture_coordinates.y, tri.b.texture_coordinates.y, tri.c.texture_coordinates.y };

    const fixed distance_inv = fixed(1) / (right - left);
    const fixed u_delta = (u_right - u_left) * distance_inv;
    const fixed v_delta = (v_right - v_left) * distance_inv;

    fixed u = u_left;
    fixed v = v_left;

    for (fixed x = left; x < right; x += 1)
    {
      // going from bottom to top, multiply y by -1
      screen.write(x, -y + sorted_tri.c.position.y, texture.read_uv(u, v));
      u += u_delta;
      v += v_delta;
    }
  }
}

}
